/*=============================================================================
	PartialHeader.h:
	This is not supposed to build and should only be used for learning purposes
=============================================================================*/

//Do remove debugging instructions
//#define CACUS_DEBUG

//These variables should persist, so we can't really afford to let the garbage collector to mess this up
static Native			ExecUndefined;
static Native			OldGetMapName; //Keeping this here...
static Native			OldPlayerCanSeeMe;
static Native			OldDynArrayElement;
static TArray<BYTE>		OldGetWeapon = TArray<BYTE>(0);
static TArray<BYTE>		OldPreLogin = TArray<BYTE>(0);
static TArray<BYTE>		OldAdminLogin = TArray<BYTE>(0);
static TArray<BYTE>		OldSetEnemy = TArray<BYTE>(0);
static TArray<BYTE>		OldRocketTick = TArray<BYTE>(0);
static TArray<BYTE>		OldSetOrders = TArray<BYTE>(0);
static TArray<BYTE>		OldShowPath = TArray<BYTE>(0);
static TArray<BYTE>		OldShowInventory = TArray<BYTE>(0);
static FCollisionCacus*	CollisionCacus = NULL;
static UObject*			NewObjects[128] = {NULL};
static INT				iNewObj = 0;
static INT				LastMapIdx = 0;
static FString			MapCachedPrefix = FString();
static TArray<FString>	MapCache = TArray<FString>(0);

inline UBOOL ValidCollidingActor( AActor* Actor)
{	return Actor->bCollideActors && !Actor->bDeleteMe;
}

inline UBOOL BrushMover( AActor *Actor)
{
	return (Actor->Brush && Actor->bIsMover );
}

inline FLOAT Sign( const FLOAT& X)
{
	if ( X > 0 )			return 1.f;
	else if ( X < 0 )		return -1.f;
	return 0.f;
}

struct FEncroachActorList : public FIteratorActorList
{
	// Variables.
	INT			OldTag;		// Old tag in octree.

	// Functions.
	FEncroachActorList( FEncroachActorList* InNext=NULL, AActor* InActor=NULL )
	:	FIteratorActorList( InNext, InActor )
	{
		if ( InActor )
			OldTag = InActor->CollisionTag;
	}
	FEncroachActorList*& GetNext()
		{ return *(FEncroachActorList**)&Next; }
};

class FPrecomputedRay
{
public:
	FVector Org;
	FVector End;
	FVector Dir;
	FVector Inv;
	FVector XAxis; //For CylCyl checks
	FVector YAxis;
	FLOAT Dist;

	FPrecomputedRay()
	{};
	FPrecomputedRay( FVector A, FVector B)
	:	End( A)
	,	Org( B)
	,	Dir( End-Org)
	{
		Dist = Dir.Size();
		if ( !FVector(Dir.X,Dir.Y,0.f).IsNearlyZero() )
		{
			XAxis = _UnsafeNormal2D( Dir.X, Dir.Y);
			YAxis = FVector( -XAxis.Y, XAxis.X, 0.f); //Rotate 90 to right in 2D
		}
		else
			appMemzero( &XAxis, 24);
		Dir = _UnsafeNormal( Dir);
		Inv = FVector(1.f/Dir.X, 1.f/Dir.Y, 1.f/Dir.Z);
	}

	inline UBOOL FitsBox( const FVector& Sample, const FBox& Container)
	{
		return (Sample.X >= Container.Min.X) && (Sample.Y >= Container.Min.Y) && (Sample.Z >= Container.Min.Z)
			&& (Sample.X <= Container.Max.X) && (Sample.Y <= Container.Max.Y) && (Sample.Z <= Container.Max.Z);
	}

	inline INT IntersectsBox( const FBox& Box)
	{
		// Quick rejects, optimized for movement code.
		if( (Org.X>Box.Max.X && End.X>Box.Max.X )
			|| ( Org.X<Box.Min.X && End.X<Box.Min.X )
			|| ( Org.Y>Box.Max.Y && End.Y>Box.Max.Y )
			|| ( Org.Y<Box.Min.Y && End.Y<Box.Min.Y )
			|| ( Org.Z>Box.Max.Z && End.Z>Box.Max.Z )
			|| ( Org.Z<Box.Min.Z && End.Z<Box.Min.Z ) )
			return 0;

		FVector T1( (Box.Min - Org) * Inv);
		FVector T2( (Box.Max - Org) * Inv);
		FLOAT TMin = ::Max(::Max(::Min(T1.X, T2.X), ::Min(T1.Y, T2.Y)), ::Min(T1.Z, T2.Z));
		FLOAT TMax = ::Min(::Min(::Max(T1.X, T2.X), ::Max(T1.Y, T2.Y)), ::Max(T1.Z, T2.Z));

		return TMax >= TMin;
	}
	UBOOL HitsVCylActor( AActor* Actor, const FVector& Extent, FVector* HitLocation, FVector* HitNormal);
	UBOOL HitsCylActor( AActor* Actor, const FVector& Extent, FVector* HitLocation, FVector* HitNormal);
};



#define MAX_OCTREE_DEPTH 6
#define OCTREE_GRID_ARRAY 7
class XC_ENGINE_API FCollisionOctree
{
public:
	FCollisionCacus* Master;
	FCollisionOctree* Parent;
	FCollisionOctree* Children[8]; //X,Y,Z based bitwise.
	FCollisionOctree* ReducedChildren[8]; //For fast operations
	FBox BoundingBox;
	FVector CenterPoint;
	BYTE CurDepth;
	BYTE ChildIndex;
	BYTE ChildCount;
	BYTE CleanupTimer;
	BYTE XPos, YPos, ZPos;
	INT MIndex; //Index on master array
	TArray<AActor*> Actors;
	TArray<AActor*> BrushMovers; //Brush && bIsMover
	//FCollisionCacus uses this
	FCollisionOctree()
	:	CurDepth(0)
	,	ChildIndex(0)
	,	ChildCount(0)
	,	CleanupTimer(20)
	,	Actors(0)
	,	BrushMovers(0)
	,	CenterPoint(0,0,0)
	,	XPos(0)	,	YPos(0)	,	ZPos(0)
	,	BoundingBox( FVector(-32768,-32768,-32768), FVector(32768,32768,32768))
	{
		appMemzero( &Parent, 4*17);
	}
	FCollisionOctree( FCollisionOctree* InParent, FBox InBounds, INT InDepth, BYTE InChild=0)
	:	Parent(InParent)
	,	BoundingBox(InBounds)
	,	CenterPoint( (InBounds.Min+InBounds.Max) * 0.5)
	,	CurDepth(InDepth)
	,	ChildIndex( InChild)
	,	CleanupTimer(20)
	,	ChildCount(0)
	,	XPos(0)	,	YPos(0)	,	ZPos(0)
	,	Actors(0)
	,	BrushMovers(0)
	{
		appMemzero( &Children[0], 4*16 ); //Both child arrays
#ifdef CACUS_DEBUG
		guard( FCollisionOctree::Constructor );
#endif
		if ( InParent )
		{
#ifdef CACUS_DEBUG
			check( InParent->ChildCount < 8 );
			check( !InParent->ReducedChildren[InParent->ChildCount] );
#endif
			InParent->ReducedChildren[InParent->ChildCount] = this;
			InParent->ChildCount++;
		}
#ifdef CACUS_DEBUG
		check( InBounds.Min.X < InBounds.Max.X );
		check( InBounds.Min.Y < InBounds.Max.Y );
		check( InBounds.Min.Z < InBounds.Max.Z );
//		debugf( NAME_Log, TEXT("[XC_GE] Creating new Octree node, Depth %i"), CurDepth);
		unguard;
#endif
	}
	~FCollisionOctree();

	void Cleanup();
	FBox GetSubOctantBox( const FVector& Point, INT& Idx);		//Fast, does not do sanity checks
	INT GetSubOctantIdx( const FVector& Point);				//Same
	UBOOL BoxesTouching( const FBox& A, const FBox& B);
	FCollisionOctree* SlowRemoveActor( AActor* Actor);	//Last resort usage, when an actor is moved without notifying the octree
	FIteratorActorList* RadiusQuery( FMemStack& Mem, const FVector& Location, const FLOAT& Radius, const DWORD& ExtraNodeFlags, FIteratorActorList* PrevLink );
	FEncroachActorList* EncroachmentQuery( FMemStack& Mem, const FBox& Box, FEncroachActorList* PrevLink); //Encroachment hack
	FCheckResult* PointQuery( FMemStack& Mem, const FVector& Location, const FVector& Extent, const DWORD& ExtraNodeFlags, FCheckResult* PrevLink );
	FCheckResult* FastPointQuery( FMemStack& Mem, const FVector& Location, const FVector& Extent, const DWORD& ExtraNodeFlags, FCheckResult* PrevLink);
	FCheckResult* PEncroachQuery( FMemStack& Mem, AActor* Actor, const FVector& Location, FCheckResult* PrevLink);
	FCheckResult* LineQuery
	(
	 FMemStack& Mem
	, const FVector& End
	, const FVector& Start
	, const FVector& Extent
	, const BYTE& ExtraNodeFlags
	, FPrecomputedRay& Ray
	,  FCheckResult* PrevLink
	);
	FCheckResult* FastLineQuery
	(
	 FMemStack& Mem
	, const FVector& End
	, const FVector& Start
	, const FVector& Extent
	, const BYTE& ExtraNodeFlags
	, FPrecomputedRay& Ray
	, FCheckResult* PrevLink
	);
	inline FBox ActorBoundingBox( AActor* Actor)				//Does not do Mover checks, fast
	{
		FVector Extent = Actor->GetCylinderExtent();
		return FBox( Actor->Location-Extent, Actor->Location+Extent);
	}
	inline UBOOL FitsBox( const FBox& Sample, const FBox& Container)
	{
		return (Sample.Min.X >= Container.Min.X) && (Sample.Min.Y >= Container.Min.Y) && (Sample.Min.Z >= Container.Min.Z)
			&& (Sample.Max.X < Container.Max.X) && (Sample.Max.Y < Container.Max.Y) && (Sample.Max.Z < Container.Max.Z);
	}
	inline UBOOL FitsBox( const FVector& Sample, const FBox& Container)
	{
		return (Sample.X >= Container.Min.X) && (Sample.Y >= Container.Min.Y) && (Sample.Z >= Container.Min.Z)
			&& (Sample.X < Container.Max.X) && (Sample.Y < Container.Max.Y) && (Sample.Z < Container.Max.Z);
	}
	inline UBOOL FitsBox( AActor* Actor, const FBox& Container)
	{
		return (Actor->Location.X - Actor->CollisionRadius >= Container.Min.X) && (Actor->Location.Y - Actor->CollisionRadius >= Container.Min.Y) && (Actor->Location.Z - Actor->CollisionHeight >= Container.Min.Z)
			&& (Actor->Location.X + Actor->CollisionRadius < Container.Max.X) && (Actor->Location.Y + Actor->CollisionRadius < Container.Max.Y) && (Actor->Location.Z + Actor->CollisionHeight < Container.Max.Z);
	}
	inline UBOOL TouchesBox( AActor* Actor, const FBox& Box)
	{
		return (Actor->Location.X + Actor->CollisionRadius >= Box.Min.X) && (Actor->Location.Y + Actor->CollisionRadius >= Box.Min.Y) && (Actor->Location.Z + Actor->CollisionHeight >= Box.Min.Z)
			&& (Actor->Location.X - Actor->CollisionRadius < Box.Max.X) && (Actor->Location.Y - Actor->CollisionRadius < Box.Max.Y) && (Actor->Location.Z - Actor->CollisionHeight < Box.Max.Z);
	}
	inline FBox ExpandByExtent( const FBox& Box, const FVector& Extent)
	{	return FBox( Box.Min-Extent, Box.Max+Extent);	}
	//Quick reject, actor should not go into a sub-box
	inline UBOOL HitsCenterAxes( AActor* Actor)
	{
		return ( ((Actor->Location.X - Actor->CollisionRadius < CenterPoint.X)
			&& (Actor->Location.X + Actor->CollisionRadius > CenterPoint.X))
		||	((Actor->Location.Y - Actor->CollisionRadius < CenterPoint.Y)
			&& (Actor->Location.Y + Actor->CollisionRadius > CenterPoint.Y))
		||	((Actor->Location.Z - Actor->CollisionHeight < CenterPoint.Z)
			&& (Actor->Location.Z + Actor->CollisionHeight > CenterPoint.Z)) );
	}
	inline UBOOL HitsCenterAxes( const FBox& Box)
	{
		return ( ((Box.Min.X < CenterPoint.X) && (Box.Max.X > CenterPoint.X))
		||	((Box.Min.Y < CenterPoint.Y) && (Box.Max.Y > CenterPoint.Y))
		||	((Box.Min.Z < CenterPoint.Z) && (Box.Max.Z > CenterPoint.Z)) );
	}
};

//====================================
// Three dimension grid
//==========
class XC_ENGINE_API FNodeGrid3 : protected TArray<FCollisionOctree*>
{
public:
	INT GridSize;
	FVector Substract; //Substract this when calculating grid position, should be identical in all grids?
	FVector BoxSize; //Dividing a vector by this (post substraction) will give coordinates!

	FNodeGrid3()
	:	TArray<FCollisionOctree*>()
	,	GridSize(0)
	{};
	INT Num() const
	{
		return ArrayNum;
	}
	void Empty( INT Slack=0 )
	{
		TArray<FCollisionOctree*>::Empty( Slack );
	}
	void SetBaseGrid( FCollisionOctree* Base)
	{
		AddItem( Base);
		Shrink();
		GridSize = 1;
		BoxSize = Base->BoundingBox.Max - Base->BoundingBox.Min;
		Substract = Base->BoundingBox.Min;
	}
	void SetGrid( INT NewGridSize, const FBox& Bounds)
	{
		check( NewGridSize > 0 );
		check( Bounds.IsValid);

		GridSize = NewGridSize;
		if ( ArrayNum > 0 )
			Empty();
		AddZeroed( NewGridSize * NewGridSize * NewGridSize);
		Shrink();

		Substract = Bounds.Min;
		BoxSize = (Bounds.Max - Bounds.Min) / (FLOAT)NewGridSize;
	}
//==Accessors==//
	const FCollisionOctree* Get( INT x, INT y, INT z ) const
	{
		return ((FCollisionOctree**)Data)[x + y*GridSize + z*GridSize*GridSize];
	}
	void Set( FCollisionOctree* In)
	{
		FVector Pos = (In->CenterPoint - Substract);
		Pos /= BoxSize;
		INT X = Clamp( (INT)Pos.X, 0, GridSize-1);
		INT Y = Clamp( (INT)Pos.Y, 0, GridSize-1);
		INT Z = Clamp( (INT)Pos.Z, 0, GridSize-1);
		In->XPos = X;		In->YPos = Y;		In->ZPos = Z;
		INT Point = X + Y * GridSize + Z * GridSize * GridSize;
		((FCollisionOctree**) Data)[Point] = In;
//		debugf( NAME_Log, TEXT("[XCGE] Added to (%i,%i,%i): %i | GridSize: %i | ArrayNum %i"), X, Y, Z, Point, GridSize, ArrayNum);
	}
	void Remove( FCollisionOctree* In)
	{
		if ( !ArrayNum ) //Something destroyed this first
			return;
		INT X = In->XPos;		INT Y = In->YPos;		INT Z = In->ZPos;
		INT Point = X + Y * GridSize + Z * GridSize * GridSize;
//		debugf( NAME_Log, TEXT("[XCGE] Removed from (%i,%i,%i): %i | GridSize: %i | ArrayNum %i"), X, Y, Z, Point, GridSize, ArrayNum);
//		check ( ((FCollisionOctree**) Data)[Point] == In );
		((FCollisionOctree**) Data)[Point] = NULL;
	}
	FCollisionOctree* Get( FVector Pos)
	{
		Pos -= Substract;
		Pos /= BoxSize;
		INT Point = Clamp( (INT)Pos.X, 0, GridSize-1) + Clamp( (INT)Pos.Y, 0, GridSize-1) * GridSize + Clamp( (INT)Pos.Z, 0, GridSize-1) * GridSize * GridSize;
		return ((FCollisionOctree**) Data)[Point];
	}
	const FCollisionOctree* Get( FVector Pos) const
	{
		Pos -= Substract;
		Pos /= BoxSize;
		INT Point = Clamp( (INT)Pos.X, 0, GridSize-1) + Clamp( (INT)Pos.Y, 0, GridSize-1) * GridSize + Clamp( (INT)Pos.Z, 0, GridSize-1) * GridSize * GridSize;
		return ((FCollisionOctree**) Data)[Point];
	}

};

class XC_ENGINE_API FCollisionCacus : public FCollisionHashBase , public FCollisionOctree
{
public:
	FCollisionHashBase* Hooked;
#ifdef CACUS_DEBUG
	INT EncroachmentQueries;
	INT PointQueries;
	INT LineQueries;
	INT FastLineQueries;
	INT RadiusQueries;
	INT DebugCycles;
#endif
	INT MinAvailable;
	TArray<FCollisionOctree*> ColList;
	AActor* JustRemoved;
	FNodeGrid3 DepthGrid[OCTREE_GRID_ARRAY]; //Zero isn't used, we're doing this for quick CurDepth access

	FCollisionCacus()
	:	ColList(0)
#ifdef CACUS_DEBUG
	,	EncroachmentQueries(0)
	,	PointQueries(0)
	,	LineQueries(0)
	,	FastLineQueries(0)
	,	RadiusQueries(0)
	,	DebugCycles(0)
#endif
	,	MinAvailable(0)
	{
		ColList.AddItem( this);
		MIndex = 0;
		Master = this;
		appMemzero( &DepthGrid, sizeof(class FNodeGrid3) * OCTREE_GRID_ARRAY );
	};
	~FCollisionCacus();

	void SetupGrids()
	{
		DepthGrid[0].SetBaseGrid( this);
		for ( INT i=1 ; i<OCTREE_GRID_ARRAY ; i++ )
			DepthGrid[i].SetGrid( 1 << i, BoundingBox);
	}

	void ChildSet( FCollisionOctree* Child);
	void Tick();
	void AddActor( AActor *Actor );
	void RemoveActor( AActor *Actor );
	FCheckResult* ActorLineCheck( FMemStack& Mem, FVector End, FVector Start, FVector Extent, BYTE ExtraNodeFlags );
	FCheckResult* ActorPointCheck( FMemStack& Mem, FVector Location, FVector Extent, DWORD ExtraNodeFlags );
	FCheckResult* ActorRadiusCheck( FMemStack& Mem, FVector Location, FLOAT Radius, DWORD ExtraNodeFlags );
	FCheckResult* ActorEncroachmentCheck( FMemStack& Mem, AActor* Actor, FVector Location, FRotator Rotation, DWORD ExtraNodeFlags );
	void CheckActorNotReferenced( AActor* Actor );
};


struct UHack_eventPreLoginHook_Parms
{
    FString Options;
    FString Address;
    FString Error;
    FString FailCode;
};

struct UHack_eventAdminLoginHook_Parms
{
	class APlayerPawn* P;
	BITFIELD ReturnValue;
};

struct MappedPropertyBlock
{
	INT Idx; //Position in array
	INT ParentIdx;
	INT TotalProps; //Faster accessing
	UClass* Class; //Actor class
	TArray< UProperty* > Properties;
	TArray< INT > ChildrenIdx;

	~MappedPropertyBlock()
	{
		if ( Properties.Num() > 0 )
			Properties.Empty();
		if ( ChildrenIdx.Num() > 0 )
			ChildrenIdx.Empty();
	}
};

//Used when for inventory/pawn travel
class XC_ENGINE_API FTravelClassMap : public TArray< MappedPropertyBlock >
{
	public:
	INT RegisterClass( UClass* InClass); //Return index in map
};

//Used for transient Actor reference cleanup
class XC_ENGINE_API FTransientClassMap : public TArray< MappedPropertyBlock >
{
	public:
	FTransientClassMap( INT i=0)
	:	TArray< MappedPropertyBlock>(0)
	{};
	INT RegisterClass( UClass* InClass);
};


class XC_ENGINE_API UXC_TravelManager : public USubsystem
{
	public:
	UXC_GameEngine*	Engine;
	BITFIELD		bAutoMode:1 GCC_PACK(4); //Do stuff myself
	BITFIELD		bNoTick:1; //Do not operate during this level
	FLOAT			TimerCounter GCC_PACK(4); //For automatic manager
	TMap_XC<INT,FString>	PlayerIdMap;
	TMap_XC<FString,FString>	TravelMap;
	FTravelClassMap	TravelClassMap; //Do not serialize, we want FLUSH and mapswitches to clear this
	
	//UObject interface
	void Serialize( FArchive& Ar );

	//USubsystem interface
	void Tick( FLOAT DeltaTime);

	//FExec interface
	UBOOL Exec( const TCHAR* Cmd, FOutputDevice& Ar );
	
	//UXC_TravelManager interface
	void FixUpPlayerId( APlayerPawn* P); //Checks for name changes
	void GenerateTravelInfo( APlayerPawn* P);
	void PostMapChange( ULevel* NewLevel);
	
	NO_DEFAULT_CONSTRUCTOR(UXC_TravelManager)
	DECLARE_CLASS(UXC_TravelManager,USubsystem,CLASS_Transient,XC_Engine)
};


class AXC_DummyPawn : private APawn
{
	friend class UXC_GameEngine;
	
	public:
	DECLARE_FUNCTION(NewPlayerCanSeeMe);
	DECLARE_FUNCTION(NewDynArrayElement);
	DECLARE_FUNCTION(NewDynArrayLength);
	DECLARE_FUNCTION(NewDynArrayInsert);
	DECLARE_FUNCTION(NewDynArrayRemove);
	DECLARE_FUNCTION(NewAppSeconds);
	DECLARE_FUNCTION(NewMakeColor);
    DECLARE_FUNCTION(PreLoginNative); //3550
    DECLARE_FUNCTION(AdminLoginNative); //3551
	//TODO:
	//native(601) final function Class<Object> GetParentClass( Class<Object> ObjClass );
	//native(1718) final function bool AddToPackagesMap( optional string PackageName );
	/* Lower-case some text */
	//static native(238) final function string Locs( string InStr );
	protected:
	UBOOL PCanSee( APlayerPawn* Other);

//    DECLARE_CLASS(APawn,AActor,0|CLASS_Abstract,XC_Engine)
};