/*=============================================================================
	Advanced travel manager functions
	By Higor, feel free to use this code on your project.
=============================================================================*/

#include "XC_Engine.h"

//Expose properties to UnrealScript
void UXC_TravelManager::StaticConstructor()
{
	guard(UXC_TravelManager::StaticConstructor);

	UClass* TheClass = GetClass();

	//I set this property to CPF_Native... maybe it's not safe to do so...
	new( TheClass, TEXT("Engine"), RF_Public) UObjectProperty( CPP_PROPERTY(Engine), TEXT("XC_TravelManager"), CPF_Native|CPF_Edit|CPF_EditConst , UXC_GameEngine::StaticClass() );
	//This should be dictated by something else, and properly updated frequently (?)
	new( TheClass, TEXT("bAutoMode"), RF_Public) UBoolProperty( CPP_PROPERTY(bAutoMode), TEXT("XC_TravelManager"), CPF_Native|CPF_Edit|CPF_EditConst );

	unguard;
}


void UXC_TravelManager::Serialize( FArchive& Ar )
{
	guard(UXC_TravelManager::Serialize);
	Super::Serialize( Ar );
	Ar << Engine << bNoTick << TravelMap;
	unguard;
}

void UXC_TravelManager::Tick( FLOAT DeltaTime)
{
	if ( bNoTick )
		return;

	ULevel* Lev = GetActiveLevel( Engine);
	if ( Engine->Client && Engine->Client->Viewports.Num() && Engine->Client->Viewports(0) && Engine->Client->Viewports(0)->Actor ) //Listen server
		FixUpPlayerId( Engine->Client->Viewports(0)->Actor ); //Local player
	if ( Lev && Lev->NetDriver && Lev->NetDriver->ClientConnections.Num() ) //Clients
	{
		for ( INT ip=0 ; ip < Lev->NetDriver->ClientConnections.Num() ; ip++ )
			if ( Lev->NetDriver->ClientConnections(ip) && Lev->NetDriver->ClientConnections(ip)->Actor )
				FixUpPlayerId( Lev->NetDriver->ClientConnections(ip)->Actor );
	}

	if ( bAutoMode ) //Timed every 1-2 seconds to scan players and update their stuff
	{
		TimerCounter -= DeltaTime;
		if ( TimerCounter < 0 )
		{
			if ( Engine->Client && Engine->Client->Viewports.Num() && Engine->Client->Viewports(0) && Engine->Client->Viewports(0)->Actor ) //Listen server
				GenerateTravelInfo( Engine->Client->Viewports(0)->Actor ); //Local player
			if ( Lev && Lev->NetDriver && Lev->NetDriver->ClientConnections.Num() ) //Clients
			{
				for ( INT ip=0 ; ip < Lev->NetDriver->ClientConnections.Num() ; ip++ )
					if ( Lev->NetDriver->ClientConnections(ip) && Lev->NetDriver->ClientConnections(ip)->Actor )
						GenerateTravelInfo( Lev->NetDriver->ClientConnections(ip)->Actor );
			}

			
			TimerCounter = 2.0;
		}
	}
}

void UXC_TravelManager::FixUpPlayerId( APlayerPawn* P)
{
	guard(UXC_TravelManager::FixUpPlayerId);
	check( P);
	if ( P->PlayerReplicationInfo )
	{
		FString* Found = NULL;
		Found = PlayerIdMap.Find( P->PlayerReplicationInfo->PlayerID );
		if ( Found ) //PlayerID in map
		{
			if ( (appStrcmp( *(P->PlayerReplicationInfo->PlayerName), **Found) != 0) )  //Name mismatch
			{
				if ( PlayerIdMap.ValueExists( P->PlayerReplicationInfo->PlayerName) ) //New name taken by another entry
					P->PlayerReplicationInfo->PlayerName = *Found;
				else //New name not taken, update this table and the main map as well
				{
					INT j;
					//Item table key name must be edited here
					for ( j=0 ; j<TravelMap.Num() ; j++ )
						if ( appStrcmp( *TravelMap.Key(j), **Found) == 0 )
						{
							TravelMap.SetKey(j, P->PlayerReplicationInfo->PlayerName);
							break;
						}
					//Now modify level table player name
					TMap_XC<FString,FString>* TmpMap = (TMap_XC<FString,FString>*) &GetActiveLevel( Engine)->TravelInfo;
					for ( j=0 ; j<TmpMap->Num() ; j++ )
						if ( appStrcmp( *TmpMap->Key(j), **Found) == 0 )
						{
							TmpMap->SetKey( j, P->PlayerReplicationInfo->PlayerName);
							break;
						}
					*Found = P->PlayerReplicationInfo->PlayerName;
				}
			}
		}
		else //Not in map, let's add
		{
			for ( INT j=0 ; j<PlayerIdMap.Num() ; j++ ) //There is a matching value with an old ID
				if ( appStrcmp( *PlayerIdMap.Value(j), *P->PlayerReplicationInfo->PlayerName) == 0 )
				{
					PlayerIdMap.SetKey( j, P->PlayerReplicationInfo->PlayerID);
					return;
				}
			PlayerIdMap.Set( P->PlayerReplicationInfo->PlayerID, *P->PlayerReplicationInfo->PlayerName );
		}
	}
	unguard;
}

void UXC_TravelManager::GenerateTravelInfo( APlayerPawn* P)
{
	guard(UXC_TravelManager::GenerateTravelInfo);
	FString STR = FString();
	FString NewLine = FString::Chr(0x0D) + FString::Chr(0x0A);
	TCHAR Temp[256] = TEXT("");
	for ( AActor* A = P ; A ; A = A->Inventory )
	{
		if ( A->bTravel )
		{
			INT Idx = TravelClassMap.RegisterClass( A->GetClass() );
			STR += FString::Printf(TEXT("Class=%s.%s Name=%s%s{%s"), A->GetClass()->GetOuter()->GetName(), A->GetClass()->GetName(), A->GetName(), *NewLine, *NewLine );
			while ( Idx >= 0 )
			{
				for ( int i=0 ; i<TravelClassMap(Idx).Properties.Num() ; i++ )
				{
					UProperty* Prop = TravelClassMap(Idx).Properties(i);
//	debugf( NAME_Log, TEXT("[PROP] %s %i %i %s"), A->GetClass()->GetName(),Idx, i, Prop->GetName() );
					check( Prop);

					if ( Prop->Matches( A, A->GetClass()->GetDefaultActor(), 0) ) //Same as default, disregard ||TREAT AS ARRAY (3RD PARAMETER)
						continue;
						//Should be treating this as array!!!
					guard(TEST2);
					Prop->ExportText( 0, Temp, (BYTE*)A,(BYTE*)A, PPF_Localized );
					unguard;
					if ( Prop->IsA( UObjectProperty::StaticClass() ) && !Prop->Matches( A, 0, 0) )
					{
						UObject** TestObj = (UObject**)  ( DWORD(A) + Prop->Offset );
//						debugf( NAME_Log, TEXT("[TRAVEL] %s"), (*TestObj)->GetName() );
						STR += FString::Printf( TEXT("%s=%s%s%s%s"), Prop->GetName(), (*TestObj)->GetName(), *NewLine, Temp, *NewLine);
					}
					else
						STR += FString::Printf(  TEXT("%s=%s%s"), Prop->GetName(), Temp, *NewLine);
				}
				Idx = TravelClassMap(Idx).ParentIdx;
			}
			STR += FString::Printf( TEXT("}%s"), *NewLine);
		}
	}
//	debugf( NAME_Log, TEXT("[TRAVEL] %s"), *STR);

	FString* Found = TravelMap.Find( P->PlayerReplicationInfo->PlayerName );
	if ( Found )
		*Found = STR;
	else
		TravelMap.Set( *P->PlayerReplicationInfo->PlayerName, *STR);
	Found = GetActiveLevel( Engine)->TravelInfo.Find( P->PlayerReplicationInfo->PlayerName );
	if ( Found )
		*Found = STR;
	else
		GetActiveLevel( Engine)->TravelInfo.Set( *P->PlayerReplicationInfo->PlayerName, *STR);
	unguard;
}

void UXC_TravelManager::PostMapChange( ULevel* NewLevel)
{
	PlayerIdMap.Empty();
	TravelMap.Empty();
	TimerCounter = 0;

	bNoTick = 0;
	if ( !NewLevel || !NewLevel->IsServer() || !NewLevel->Actors.Num() || !NewLevel->NetDriver )
		bNoTick = 1;
}

INT FTravelClassMap::RegisterClass( UClass* InClass) //Return index in map
{
	guard(FTravelClassMap::RegisterClass);
	if ( !InClass->IsChildOf( AActor::StaticClass()) )
		return -1;
	if ( !ArrayNum )
	{
		AddZeroed();
		(*this)(0).ParentIdx = -1;
		(*this)(0).Class = AActor::StaticClass();
		if ( InClass == AActor::StaticClass() )
			return 0;
	}
	//Find direct subclasses, register whatever's necessary
	UClass* Parent = AActor::StaticClass();
	INT ParentIdx = 0;
	while ( Parent )
	{
		UBOOL Found = 0;
		UClass* CurClass = InClass;
		if ( CurClass == Parent ) //Cur has been added!
			break;
		while ( CurClass->GetSuperClass() != Parent )
			CurClass = CurClass->GetSuperClass();
		//This class's parent is obviously on this list, attempt to find using parent's children array
		for ( int i=0 ; i< (*this)(ParentIdx).ChildrenIdx.Num() ; i++ )
		{
			if ( (*this)( (*this)(ParentIdx).ChildrenIdx(i) ).Class == CurClass ) //This class is listed on parent's children, do not register
			{
				Found = 1;
				ParentIdx = (*this)(ParentIdx).ChildrenIdx(i);
				break;
			}
		}
		//Register this class if necessary
		if ( !Found )
		{
			AddZeroed();
			Last().Class = CurClass;
			Last().Idx = ArrayNum - 1;
			Last().ParentIdx = ParentIdx;
			(*this)(ParentIdx).ChildrenIdx.Add();
			(*this)(ParentIdx).ChildrenIdx.Last() = Last().Idx; //Add the new child IDX (this) onto the parent
			
			UField* Field = CurClass->Children;
			while ( Field ) //Find Travel properties
			{
				if ( Field->IsA( UProperty::StaticClass()) && (((UProperty*)Field)->PropertyFlags & CPF_Travel) ) //Travel property, list it
				{
					Last().Properties.AddZeroed();
					Last().Properties.Last() = (UProperty*) Field;
				}
				Field = Field->Next;
			}
			ParentIdx = Last().Idx;
		}
		Parent = CurClass;
	}
	return ParentIdx;
	unguard;
}


/*-----------------------------------------------------------------------------
	The End.
-----------------------------------------------------------------------------*/
