//=============================================================================
// XC_EditorLoader.
// Editor entry point for XC_Engine.
//=============================================================================
class XC_EditorLoader expands Object
	native;


//////////////////////////////////////////////////////////
/*  XC_Engine specific UnrealScript features.
//////////////////////////////////////////////////////////

=================
Native functions.
Your class/package doesn't need to be native to be able to use these opcodes.
You may define these opcodes anywhere in your classes, just declare the functions you want to use.
On the Array_ opcodes, you may replace 'Type' with absolutely ANY datatype of your choice.

If you want to make your code compatible with both UT99 modes, add IF checks that allow the game to
differ when to call the slow function and when to use the native opcodes.
Differing between 227 and UT99 (with and without XCGE) takes a whole different check set I won't address here.

Just a reminder, these opcodes won't be availabe during online play, so attempting to run said
code while connected to a server will cause a crash.

***** XC_GameEngine only opcodes
- (3550) PreLoginHook. For internal use, do not call.
- (3551) AdminLoginHook. For internal use, do not call.

***** 227 Based opcodes
***** Compiled code with these opcodes is compatible with Unreal 227 and vice-versa.
- (10) Dynamic Array accessor [], doesn't need to be predefined.
- native(198) static final function color MakeColor( byte R, byte G, byte B, optional byte A );
- native(640) static final function int Array_Length_Type( out array<Type> Ar, optional int SetSize);
- native(641) static final function bool Array_Insert_Type( out array<Type> Ar, int Offset, optional int Count );
- native(642) static final function bool Array_Remove_Type( out array<Type> Ar, int Offset, optional int Count );
- native(643) static final function float AppSeconds();


-------
-- SIMPLE EXAMPLE OF CODE THAT RUNS IN ANY UT99 MODE:
var() const bool bTrue;	//This is set to True all the time, used for multilanguage support
var bool bXCGE;
...
	// Define this once at object creation, or everytime the level changes (if used from UWindow)
	bXCGE = ConsoleCommand("Get ini:Engine.Engine.GameEngine bHackedEngine") ~= GetPropertyText("True");
...

	if ( bXCGE )	//Execute XC_Engine specific code here
		Canvas.DrawColor = MakeColor( 150, 150, 0);
	else			//Fall back to default code
	{
		Canvas.DrawColor.R = 150;
		Canvas.DrawColor.G = 150;
		Canvas.DrawColor.B = 0;
	}
-------
-- PROFILING A LONG CHUNK OF CODE
var float Clock;
...
	//Start profiling here
	Clock = AppSeconds();
...
	//End profiling here
	Clock = AppSeconds() - Clock;
	Log("Time needed to execute: "$Clock);
-------

=================
Extra properties:

* Class			PropertyName	(Property Type)	(Property Flags)

- GameEngine	Level			(Level)			(const, editconst)
- GameEngine	Entry			(Level)			(const, editconst)

- DemoRecDriver	DemoFileName	(string)		(const, editconst)

- LevelBase		NetDriver		(NetDriver)		(const, editconst)
- LevelBase		DemoRecDriver	(NetDriver)		(const, editconst)
- LevelBase		Engine			(Engine)		(const, editconst)
- LevelBase		URL_Protocol	(string)		(const, editconst)
- LevelBase		URL_Host		(string)		(const, editconst)
- LevelBase		URL_Port		(int)			(const, editconst)
- LevelBase		URL_Map			(string)		(const, editconst)
- LevelBase		URL_Options		(array<string>)	(const, editconst)
- LevelBase		URL_Portal		(string) 		(const, editconst)

- NetDriver		ClientConnections	(array<NetConnection>)	(const, editconst)
- NetDriver		ServerConnection	(NetConnection)	(const, editconst)

Referencing these properties directly will prevent the package from being loadable.
Use GetPropertyText("PropName") to access these.
*/

defaultproperties
{
}
