//=============================================================================
// XC_ServerActor.
//=============================================================================
class XC_ServerActor expands Actor
	config;

struct LoginInfo
{
	var() PlayerPawn Player;
	var() float DenyUntil;
	var() int BadLoginCount;
};

var() config int MaxBadLoginAttempts;
var() config float LoginTryAgainTime;
var() config bool bKickAfterMaxLogin;

var array<LoginInfo> LInfos;

event PostBeginPlay()
{
	//Become the primary AdminLoginHook (replaces previous ones if defined)
	//Will fail if named actor doesn't have
	//event bool AdminLoginHook( PlayerPawn P)
	ConsoleCommand( "AdminLoginHook "$Name);
	//Warning, SaveConfig() will be called for this actor by XC_GameEngine once map ends!!!
}

//Return true if player is allowed to AdminLogin
event bool AdminLoginHook( PlayerPawn P)
{
	local int i, iP;

	iP = Array_Length_LI( LInfos);
	while ( i<iP )
	{
		if ( LInfos[i].Player == none || LInfos[i].Player.bDeleteMe )
		{
			Array_Remove_LI( LInfos, i);
			iP--;
			continue;
		}
		if ( LInfos[i].Player == P )
		{
			if ( (Level.TimeSeconds < LInfos[i].DenyUntil) || (LInfos[i].BadLoginCount > MaxBadLoginAttempts) ) //This player is spamming the login
			{
				if ( LInfos[i].BadLoginCount++ > MaxBadLoginAttempts && bKickAfterMaxLogin )
				{
					P.ClientMessage("Kicked due to excessive AdminLogin attempts("$MaxBadLoginAttempts$")");
					if ( ViewPort(P.Player) == none ) //Never destroy local player
						P.Destroy();
					Array_Remove_LI( LInfos, i);
					iP--;
				}
				LInfos[i].DenyUntil += LoginTryAgainTime * Level.TimeDilation * 0.1;
				return false;
			}
			LInfos[i].DenyUntil = Level.TimeSeconds + LoginTryAgainTime * Level.TimeDilation;
			return true;
		}
		i++;
	}
	Array_Insert_LI( LInfos, iP);
	LInfos[iP].Player = P;
	LInfos[iP].DenyUntil = Level.TimeSeconds + LoginTryAgainTime * Level.TimeDilation;
	return true;
}

//Gain access to these XC_GameEngine opcodes
native(640) static final function int Array_Length_LI( out array<LoginInfo> Ar, optional int SetSize);
native(641) static final function bool Array_Insert_LI( out array<LoginInfo> Ar, int Offset, optional int Count );
native(642) static final function bool Array_Remove_LI( out array<LoginInfo> Ar, int Offset, optional int Count );

defaultproperties
{
     MaxBadLoginAttempts=100
     LoginTryAgainTime=4
     bKickAfterMaxLogin=false
}
